from . import MeasureTransformation, NumpyMeasure, IdentityTransformation
from .activations_iterator import ActivationsIterator
from tmeasures import MeasureResult
from .stats_running import RunningMeanWelford
from .quotient import divide_activations
from .aggregation import DistanceAggregation
import tmeasures as tm


class TransformationDistanceInvariance(NumpyMeasure):
    def __init__(self, distance_aggregation: DistanceAggregation):
        super().__init__()
        self.distance_aggregation = distance_aggregation

    def __repr__(self):
        return f"{self.abbreviation()}(da={self.distance_aggregation})"

    def eval(self, activations_iterator: ActivationsIterator, verbose=False) -> MeasureResult:
        layer_names = activations_iterator.layer_names()
        n_intermediates = len(layer_names)
        mean_running = [RunningMeanWelford() for i in range(n_intermediates)]
        for x, transformation_activations_iterator in activations_iterator.samples_first():
            # transformation_activations_iterator can iterate over all transforms
            for x_transformed, activations in transformation_activations_iterator:
                for j, layer_activations in enumerate(activations):
                    # calculate the distance aggregation only for this batch
                    layer_measure = self.distance_aggregation.apply(layer_activations)
                    # update the mean over all transformation
                    mean_running[j].update(layer_measure)

        # calculate the final mean over all samples (and layers)
        mean_variances = [b.mean() for b in mean_running]
        return MeasureResult(mean_variances, layer_names, self)

    def name(self):
        return "Transformation Distance"

    def abbreviation(self):
        return "TD"


class SampleDistanceInvariance(NumpyMeasure):
    def __init__(self, distance_aggregation: DistanceAggregation):
        super().__init__()
        self.distance_aggregation = distance_aggregation

    def __repr__(self):
        return f"{self.abbreviation()}(da={self.distance_aggregation})"

    def eval(self, activations_iterator: ActivationsIterator, verbose=False) -> MeasureResult:
        layer_names = activations_iterator.layer_names()
        n_layers = len(layer_names)
        mean_running = [RunningMeanWelford() for i in range(n_layers)]

        for transformation, transformation_activations in activations_iterator.transformations_first():
            # calculate the variance of all samples for this transformation
            for x, batch_activations in transformation_activations:
                for j, layer_activations in enumerate(batch_activations):
                    layer_measure = self.distance_aggregation.apply(layer_activations)
                    mean_running[j].update(layer_measure)

        # calculate the final mean over all transformations (and layers)
        mean_variances = [b.mean() for b in mean_running]
        return MeasureResult(mean_variances, layer_names, self)


class NormalizedDistanceInvariance(NumpyMeasure):
    def __init__(self, distance_aggregation: DistanceAggregation,
                 pre_normalization_transformation: MeasureTransformation = IdentityTransformation()):
        self.distance_aggregation = distance_aggregation
        self.td = TransformationDistanceInvariance(distance_aggregation)
        self.sd = SampleDistanceInvariance(distance_aggregation)
        self.pre_normalization_transformation = pre_normalization_transformation

    def eval(self, activations_iterator: ActivationsIterator, verbose=False) -> MeasureResult:
        td_result = self.td.eval(activations_iterator, verbose)
        sd_result = self.sd.eval(activations_iterator, verbose)

        td_result = self.pre_normalization_transformation.apply(td_result)
        sd_result = self.pre_normalization_transformation.apply(sd_result)

        result = divide_activations(td_result.layers, sd_result.layers)
        return MeasureResult(result, activations_iterator.layer_names(), self)

    def __repr__(self):
        return f"{self.abbreviation()}(pnt={self.pre_normalization_transformation},da={self.distance_aggregation})"



from math import floor
from typing import Dict, List
import tmeasures as tm
from torch.utils.data import Dataset

from tmeasures.pytorch.transformations import IdentityTransformationSet

from .stats_running import RunningMeanAndVarianceWelford, RunningMeanWelford
from .base import PyTorchLayerMeasure, PyTorchMeasure, PyTorchMeasureByLayer, PyTorchMeasureOptions, PyTorchMeasureResult, STMatrixIterator
from .activations_iterator import PytorchActivationsIterator
from . import ActivationsModule
from .layer_measures import Variance
from .quotient import  divide_activations
import torch


default_alpha=0.99
default_sign=1


class PercentActivationThreshold(PyTorchLayerMeasure):
    def __init__(self,sign:float=1,percent:torch.DoubleTensor=0.99) -> None:
        super().__init__()
        self.percent=percent
        self.sign=sign
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(sign={self.sign},p={self.percent})"

    def eval(self, st_iterator: STMatrixIterator, layer_name: str):
        n = len(st_iterator)
        values = torch.Tensor()
        for row, row_iterator in enumerate(st_iterator):
            if row == 0:
                m = len(row_iterator)
            for batch_i, batch_activations in enumerate(row_iterator):
                if row == 0 and batch_i == 0:
                    # initialize matrix to store activation values 
                    # One row for each activation
                    values = torch.zeros((n,m,*batch_activations.shape))
                    
                values[row,batch_i,:] = batch_activations*self.sign
        # change shape of values to Activations x Samples
        # where each row has all the samples of a particular activation
        original_shape = values.shape[3:]
        print("before",values.shape)
        samples_n = values.shape[0]*values.shape[1]*values.shape[2]
        values = values.reshape((samples_n,-1)).transpose(1,0)
        # sort rows, one for each activation
        # print(values.shape)
        values, _ = torch.sort(values)
        # determine index of threshold
        i = floor(values.shape[1]*self.percent)
        thresholds= values[:,i]
        # reshape thresholds to original shape
        thresholds = thresholds.reshape(original_shape)
        return thresholds

class NormalPValueThreshold(PyTorchLayerMeasure):
    def __init__(self,sign:float=1,alpha:torch.DoubleTensor=0.99) -> None:
        super().__init__()
        self.alpha=alpha
        self.sign=sign

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(sign={self.sign},alpha={self.alpha})"

    def eval(self, st_iterator: STMatrixIterator, layer_name: str):
        
        mean = RunningMeanAndVarianceWelford()
        for row, row_iterator in enumerate(st_iterator):
            for batch_n, batch_activations in enumerate(row_iterator):
                mean.update_batch(batch_activations.double()*self.sign)

        
        # return mean.mean()
        μ,σ = mean.mean(),mean.std()
        original_shape = μ.shape
        
        σ[σ<1e-16]=1e-16
        d = torch.distributions.Normal(μ,σ)
        alpha = torch.tensor([self.alpha]).to(μ.device)
        p_values = d.icdf(alpha)
        
        # μ,σ = μ.reshape(μ.numel()),σ.reshape(σ.numel())
        # p_values=torch.zeros(μ.shape)
        # alpha = torch.tensor([self.alpha]).to(μ.device)
        # for j,(mu,sigma) in enumerate(zip(μ,σ)):
        #     if sigma>0:
        #         d = torch.distributions.Normal(mu,sigma)
        #         t = d.icdf(alpha)
        #     else:
        #         t=mu
        #     p_values[j]=t
        # print(f"NormalPvalue {layer_name} p_values: {p_values.shape} ")
        # p_values = p_values.reshape(original_shape)
        
        return p_values
   
class MeanFiringRate(PyTorchLayerMeasure):
    def __init__(self,sign:float,thresholds:Dict[str,torch.Tensor]) -> None:
        super().__init__()
        self.sign=sign
        self.thresholds=thresholds

    def eval(self, st_iterator: STMatrixIterator, layer_name: str):
        thresholds=self.thresholds[layer_name]
        mean = RunningMeanWelford()
        for row, row_iterator in enumerate(st_iterator):
            for batch_n, batch_activations in enumerate(row_iterator):
                if self.sign != 1:
                    batch_activations *= self.sign
                values = (batch_activations > thresholds) * 1.0
                mean.update_all(values)
        return mean.mean()


class GoodfellowInvariance(PyTorchMeasure):

    g_key="global"
    l_key="local"
    t_key= "thresholds"

    def __init__(self, sign=1, global_transformations = IdentityTransformationSet(),threshold_algorithm=NormalPValueThreshold(sign=1,alpha=0.99)):
        assert sign in [1,-1]
        super().__init__()
        self.sign=sign
        self.threshold_algorithm=threshold_algorithm
        self.global_transformations = global_transformations
        

    def eval(self, dataset: Dataset, transformations: tm.TransformationSet, model: ActivationsModule,o: PyTorchMeasureOptions):
        
        threshold_measures = PyTorchMeasureByLayer(self.threshold_algorithm)
        
        thresholds = threshold_measures.eval(dataset,self.global_transformations,model,o)
        
        mean_firing_rate = MeanFiringRate(self.sign,thresholds.layers_dict())
        
        g_result = PyTorchMeasureByLayer(mean_firing_rate).eval(dataset,self.global_transformations,model,o)
        
        l_result = PyTorchMeasureByLayer(mean_firing_rate).eval(dataset,transformations,model,o)
        
        # self.g = GoodfellowNormalGlobalInvariance(self.alpha, self.sign)
        # g_result = self.g.eval(dataset, transformations, model,o)
        # thresholds = g_result.extra_values[GoodfellowNormalGlobalInvariance.thresholds_key]
        # self.l = GoodfellowNormalLocalInvariance(thresholds, self.sign)
        # l_result = self.l.eval(dataset, transformations, model,o)

        ratio = divide_activations(l_result.layers,g_result.layers)
        extra = {self.g_key:g_result,self.l_key:l_result,self.t_key:thresholds}

        return PyTorchMeasureResult(ratio, model.activation_names(), self,extra_values=extra)

    def __repr__(self):
        return f"{self.abbreviation()}({self.threshold_algorithm})"



import numpy as np
import abc
import itertools
from typing import Iterator

class TransformationParameters(abc.ABC,list):
    def __init__(self,members:[]):
        super().__init__(members)

    @abc.abstractmethod
    def id(self)->str:
        pass

    def __repr__(self)->str:
        return self.id()

    def scale3(self, scales, include_identity=False):
        result=[]
        if include_identity:
            result.append((1,1))
        for s in scales:
            if s == 1 and include_identity:
                continue
            s2 = [(1,s),(s,1),(s,s)]
            result+=s2
        return result

    def scale1(self, scales, include_identity=False):
        result=[]
        if include_identity:
            result.append((1,1))
        result+= [(s,s) for s in scales if not (s==1 and include_identity)]
        return result

    def translations8(self, values, include_identity=False):
        translations=[]
        if include_identity:
            translations.append([0,0])

        for d in values:
            p=[0,d,-d]
            combinations = list(itertools.product(p,p))
            combinations=combinations[1:]# remove (0,0)
            translations+=combinations
        return translations

    def translations4(self, values, include_identity=False):
        translations=[]
        if include_identity:
            translations.append([0,0])

        for d in values:
            translations+=[ (0,d),(d,0),(0,-d),(-d,0)] # clockwise NESW
        return translations

    def translations4diag(self, values, include_identity=False):
        translations=[]
        if include_identity:
            translations.append([0,0])

        for d in values:
            translations+=[ (-d,-d),(-d,d),(d,d),(d,-d)] #clockwise SW NW NE SE
        return translations

class ConstantParameters(TransformationParameters):
    def __init__(self,values:[]):
        super().__init__(values)
    def id(self):
        return f"C({self[:]})"


class UniformRotation(TransformationParameters):
    def __init__(self, n:int=1,angles:float=0):
        self.n=n
        if n<0: raise ValueError(f"Wrong value for n")
        if  angles <0 or angles >1: raise ValueError(f"Wrong angle range {angles}")
        self.angles=angles
        super().__init__(self.values())

    def values(self):
        if self.n==0:
            return [0]
        else:
            # TODO change to -self.angles to self.angles and use 180 as max
            # So that changes in angle correspond more closely to changes in
            # transformation "complexity"
            # In this way, both α and -α always belong to the same t set
            a=self.angles/2
            return np.linspace(-a, a, self.n)

    def id(self):
        return f"UR({self.angles},{self.n})"

class ScaleParameters(TransformationParameters):
    pass

class ScaleUniform(ScaleParameters):
    def __init__(self, n:int=0, min_downscale:float=1, max_upscale:float=1,include_identity=True):
        # generates n*2*3+1=n*6+1 scales
        # It generates n 1D downscales uniformly sampled from (min_downscale,1)
        # and n 1D upscales uniformly sampled from (1,max_upscale)
        # for n*2 total 1D scales
        # Then each those 1D scales are transformed into 3 2D scales such that:
        # s -> [ (1,s), (s,1), (s,s)]
        # The identity transform is also included
        self.n=n
        if  min_downscale <0 or min_downscale>1 or max_upscale<1: raise ValueError(f"Values must satisfy  0<{min_downscale}<=1<={max_upscale}")
        self.min_downscale, self.max_upscale= min_downscale, max_upscale
        self.include_identity=include_identity
        super().__init__(self.values())

    def values(self) ->[float,float]:
        if self.n==0:
            return (1,1)
        upscale = np.linspace(1,self.max_upscale,self.n+1,endpoint=True)
        upscale=upscale[1:]# remove scale=1
        downscale = np.linspace(self.min_downscale,1,self.n,endpoint=False)
        return self.scale3(downscale)+self.scale3(upscale,include_identity=self.include_identity)

    def id(self):
        return f"US({self.min_downscale}-{self.max_upscale},{self.n})"

class ScaleUniformSymmetric(ScaleUniform):
    def __init__(self,n:int=0,intensity:float=0):
        super().__init__(n,1-intensity,1+intensity)

class TranslationUniform(TransformationParameters):
    def __init__(self, n:int=1,max_intensity:float=0,):
        assert max_intensity >= 0 and max_intensity <= 100
        self.n=n
        self.max_intensity=max_intensity
        super().__init__(self.values())

    def values(self):
        if self.n==0:
            return [(0,0)]
        values=np.linspace(0, self.max_intensity, self.n+1,endpoint=True)
        values=values[1:] # remove 0
        return self.translations8(values,include_identity=True)
    def id(self):
        return f"UT({self.max_intensity},{self.n})"


class NoRotation(TransformationParameters):
    def __init__(self):
        super().__init__([0])
    def id(self):
        return f"NR"

class NoScale(TransformationParameters):
    def __init__(self):
        super().__init__([(1,1)])
    def id(self):
        return f"NS"

class NoTranslation(TransformationParameters):
    def __init__(self):
        super().__init__([(0,0)])
    def id(self):
        return f"NT"



import tmeasures as tm
from tmeasures import  MeasureResult
from . import NumpyMeasure,MeasureTransformation,IdentityTransformation
from .activations_iterator import ActivationsIterator
from .stats_running import RunningMeanAndVarianceWelford, RunningMeanWelford
from .base import NumpyMeasure
from .quotient import divide_activations

class TransformationVarianceSameEquivariance(NumpyMeasure):
    def __init__(self):
        super().__init__()

    def __repr__(self):
        return f"TVSE()"

    def eval(self, activations_iterator: ActivationsIterator,verbose=False) -> MeasureResult:
        activations_iterator = activations_iterator.get_inverted_activations_iterator()

        mean_running = None
        for x,transformation_activations  in activations_iterator.samples_first():
            transformation_variances_running = None
            #calculate the running mean/variance/std over all transformations of x
            for x_transformed, activations in transformation_activations:
                if mean_running is None:
                    n_layers = len(activations)
                    mean_running = [RunningMeanWelford() for i in range(n_layers)]
                if transformation_variances_running is None:
                    n_layers = len(activations)
                    transformation_variances_running = [RunningMeanAndVarianceWelford() for i in range(n_layers)]
                for i, layer_activations in enumerate(activations):
                    # apply function to conv layers
                    # update the mean over all transformations for this sample
                    transformation_variances_running[i].update_all(layer_activations)
            # update the mean with the numpy sample of all transformations of x
            for i,layer_variance in enumerate(transformation_variances_running):
                mean_running[i].update(layer_variance.std())

        # calculate the final mean over all samples (for each layer)
        mean_variances = [b.mean() for b in mean_running]
        return MeasureResult(mean_variances, activations_iterator.layer_names(), self)


class SampleVarianceSameEquivariance(NumpyMeasure):
    def __init__(self):
        super().__init__()

    def __repr__(self):
        return f"SVSE()"

    def eval(self, activations_iterator: ActivationsIterator,verbose=False) -> MeasureResult:
        activations_iterator:ActivationsIterator = activations_iterator.get_inverted_activations_iterator()
        ts = list(map(str, (activations_iterator.get_transformations())))


        mean_variances_running = None

        for transformation, samples_activations_iterator in activations_iterator.transformations_first():
            samples_variances_running = None
            # calculate the variance of all samples for this transformation
            for x, batch_activations in samples_activations_iterator:
                if mean_variances_running is None:
                    n_layers = len(batch_activations)
                    mean_variances_running = [RunningMeanWelford() for i in range(n_layers)]
                if samples_variances_running is None:
                    n_layers = len(batch_activations)
                    samples_variances_running = [RunningMeanAndVarianceWelford() for i in range(n_layers)]
                for j, layer_activations in enumerate(batch_activations):
                    samples_variances_running[j].update_all(layer_activations)
            # update the mean over all transformation (and layers)
            for layer_mean_variances_running, layer_samples_variance_running in zip(mean_variances_running,samples_variances_running):
                layer_mean_variances_running.update(layer_samples_variance_running.std())

        # calculate the final mean over all transformations (and layers)

        mean_variances = [b.mean() for b in mean_variances_running]
        return MeasureResult(mean_variances, activations_iterator.layer_names(), self)




class NormalizedVarianceSameEquivariance(NumpyMeasure):
    def __init__(self, pre_normalization_transformation: MeasureTransformation=IdentityTransformation()):
        self.sv = SampleVarianceSameEquivariance()
        self.tv = TransformationVarianceSameEquivariance()
        self.pre_normalization_transformation = pre_normalization_transformation

    transformation_key=TransformationVarianceSameEquivariance.__name__
    sample_key=SampleVarianceSameEquivariance.__name__

    def eval(self, activations_iterator: ActivationsIterator,verbose=False) -> MeasureResult:

        sample_result=self.sv.eval(activations_iterator)
        transformation_result = self.tv.eval(activations_iterator)

        # TODO REFACTOR NEW layer_transformation.py
        transformation_result = self.pre_normalization_transformation.apply(transformation_result)
        sample_result = self.pre_normalization_transformation.apply(sample_result)

        extra_values={ self.transformation_key:transformation_result,
                       self.sample_key:sample_result,
                       }

        result=divide_activations(transformation_result.layers, sample_result.layers)
        return MeasureResult(result, transformation_result.layer_names, self,extra_values)

    def __repr__(self):
        ca = f"ca={self.pre_normalization_transformation}"

        return f"NVSE({ca})"
    def name(self):
        return "Normalized Variance Same Equivariance"
    def abbreviation(self):
        return "NVSE"


"""
The tmeasures library contain measures for neural networks internals in terms of their response to transformations
"""

import logging
logger = logging.getLogger(__name__)
log_level = logging.WARN
logger.setLevel(log_level)

# ORDER MATTERS IN THIS FILE
# IMPORT BASIC STUFF FIRST
from .transformation import TransformationSet,Transformation,IdentityTransformation,InvertibleTransformation
from .measure import MeasureResult, StratifiedMeasureResult,Measure
from . import pytorch,visualization
from . import np







import numpy as np
import abc


class RunningMeasure(abc.ABC):
    @abc.abstractmethod
    def clear(self):
        pass

    @abc.abstractmethod
    def update(self, x: np.ndarray):
        pass

    def update_all(self, x: np.ndarray):
        for i in range(x.shape[0]):
            self.update(x[i, :])

    @abc.abstractmethod
    def update_batch(self, x: np.ndarray):
        pass


class RunningMean(RunningMeasure):

    @abc.abstractmethod
    def mean(self):
        pass


class RunningVariance(RunningMeasure):
    @abc.abstractmethod
    def var(self):
        pass

    def std(self):
        return np.sqrt(self.var())


class RunningMeanSets(RunningMean):
    def __repr__(self):
        return f"RunningMeanSets(n={self.n},mean={self.mean().shape})"

    def __init__(self):
        self.clear()

    def clear(self):
        self.n = 0
        self.mu = 0

    def mean(self):
        return self.mu

    def update_batch(self, x: np.ndarray):
        m = x.shape[0]
        mu_x = x.mean(axis=0)
        n = self.n
        self.n = m + n
        if n == 0:
            self.mu = mu_x
        else:
            c1, c2 = n / self.n, m / self.n
            self.mu = c1 * self.mu + c2 * mu_x

    def update(self, x: np.ndarray):
        self.n += 1
        if self.n == 1:
            self.mu = x
        else:
            self.mu = self.mu + (x - self.mu) / self.n


# noinspection PyAttributeOutsideInit
class RunningMeanVarianceSets(RunningMean, RunningVariance):
    def __repr__(self):
        return f"RunningMeanVarianceSets(n={self.n})"

    def __init__(self):
        self.clear()

    def clear(self):
        self.n = 0
        self.mu = 0
        self.v = 0

    def update_batch(self, x: np.ndarray):
        m = x.shape[0]
        mu_x = x.mean(axis=0)
        if m==1:
            v_x=0
        else:
            # print(x)
            # import matplotlib.pyplot as plt
            # f,ax=plt.subplots(x.shape[0])
            # for i in range(x.shape[0]):
            #     im = x[i,10,:,:]
            #     # im = im.transpose((1,2,))
            #
            #     ax[i].imshow(im)
            # plt.savefig(f"batch{self.n}.png")
            # print(x[:,4:6,16,14])
            v_x = x.var(axis=0,ddof=0)
            # print(v_x[:,14:16,14:16])

        n = self.n
        self.n = m + n
        # print("pre")
        if n == 0:
            self.mu = mu_x
            # print(m,v_x)
            self.v = v_x
        else:
            c1, c2= n / self.n, m / self.n
            c3 = c1 * c2
            mu = self.mu
            self.mu = c1 * self.mu + c2 * mu_x
            self.v = c1 * self.v + c2 * v_x + c3 * ((mu - mu_x) ** 2)

    def var(self):
        return self.v*self.n/(self.n-1)

    def mean(self):
        return self.mu

    def update(self, x: np.ndarray):
        self.update_batch(x[np.newaxis,])


class RunningMeanWelford(RunningMean):
    def __repr__(self):
        return f"RunningMeanWelford(n={self.n},mean={self.mean().shape})"

    def __init__(self):
        self.clear()

    def clear(self):
        self.n = 0
        self.mu = np.zeros(1, dtype=np.float32)

    def update(self, x):
        self.n += 1

        if self.n == 1:
            self.mu = x
        else:
            self.mu = self.mu + (x - self.mu) / self.n

    def update_batch(self, x: np.ndarray):
        k = x.shape[0]
        if self.n == 0:
            self.mu = x.mean(axis=0)
            self.n = k
        else:
            self.n += k
            x_sum = x.sum(axis=0)
            self.mu = self.mu + (x_sum - self.mu * k) / self.n

    def mean(self):
        return self.mu if self.n > 0 else np.zeros(1)


class RunningMeanAndVarianceWelford(RunningMean, RunningVariance):

    def __repr__(self):
        return f"RunningMeanAndVarianceWelford(n={self.n},mean={self.mean().shape})"

    def __init__(self):
        self.clear()

    def clear(self):
        self.n = 0
        self.m = 0
        self.s = 0

    def update(self, x: np.ndarray):
        self.n += 1
        diff = x - self.m
        self.m = self.m + (diff / self.n)
        self.s = self.s + diff * (x - self.m)

    def update_batch(self, x: np.ndarray):
        # see https://stackoverflow.com/questions/56402955/whats-the-formula-for-welfords-algorithm-for-variance-std-with-batch-updates
        k = x.shape[0]

        self.n += k
        diff = x - self.m
        self.m = self.m + diff.sum(axis=0) / self.n
        # print(diff * (x - self.m))
        # print(self.m[:10,0,0])
        diff2 = (diff * (x - self.m)).sum(axis=0)
        self.s = self.s + diff2
        # print("s=",self.s.min(),self.s.max())

    def mean(self):
        return self.m if self.n > 0 else np.zeros_like(self.m)

    def var(self):
        return self.s / (self.n - 1) if self.n > 1 else np.zeros_like(self.s)




from typing import List
from ..base import NumpyMeasure
from tmeasures import MeasureResult
from ..activations_iterator import ActivationsIterator

import multiprocessing
from queue import Queue
from threading import Thread

import abc

class LayerMeasure(abc.ABC):
    def __init__(self,id:int,name:str):
        self.id=id
        self.name=name

    def eval_private(self,q:Queue,inner_q:Queue,rq:Queue):
        result=self.eval(q,inner_q)
        rq.put(result)


    @abc.abstractmethod
    def eval(self,q:Queue,inner_q:Queue):
        pass

    def queue_as_generator(self,q: Queue):
        while True:
            v = q.get()
            if v is None:
                break
            else:
                yield v

from enum import Enum

class ActivationsOrder(Enum):
    TransformationsFirst = "tf"
    SamplesFirst = "sf"

class PerLayerMeasure(NumpyMeasure, abc.ABC):

    def __init__(self,activations_order:ActivationsOrder,queue_max_size=1,multiprocess=False):
        if multiprocess:
            print("Warning, deadlocks when using multiprocess are possible")
        self.activations_order = activations_order
        self.queue_max_size=queue_max_size
        if multiprocess:
            self.process_class = multiprocessing.Process
            self.queue_class = multiprocessing.Queue
        else:
            self.process_class = Thread
            self.queue_class = Queue

    @abc.abstractmethod
    def generate_layer_measure(self, i:int, name:str) -> LayerMeasure:
        pass

    def signal_iteration_end(self, queues: List[Queue]):
        self.put_value(queues,None)

    def put_values(self,queues: List[Queue],values:List):
        for q,v in zip(queues,values):
            q.put(v)

    def put_value(self,queues:List[Queue],value):
        for q in queues:
            q.put(value)

    def start_threads(self,threads:List):
        for t in threads:
            t.start()
    def wait_for_threads(self, threads: List):
        for t in threads:
            t.join()

    def eval(self,activations_iterator:ActivationsIterator)->MeasureResult:
        names = activations_iterator.layer_names()
        layers = len(names)
        layer_measures = [self.generate_layer_measure(i, name) for i, name in enumerate(names)]

        queues = [self.queue_class(self.queue_max_size) for i in range(layers)]
        inner_queues = [self.queue_class(self.queue_max_size) for i in range(layers)]
        result_queues = [self.queue_class(self.queue_max_size) for i in range(layers)]

        threads = [self.process_class(target=c.eval_private, args=[q, qi,qr],daemon=True) for c, q, qi, qr in
                   zip(layer_measures, queues, inner_queues,result_queues )]

        self.start_threads(threads)
        if self.activations_order == ActivationsOrder.SamplesFirst:
            self.eval_samples_first(activations_iterator,queues,inner_queues)
        elif self.activations_order == ActivationsOrder.TransformationsFirst:
            self.eval_transformations_first(activations_iterator, queues, inner_queues)
        else:
            raise ValueError(f"Unknown activations order {self.activations_order}")
        self.wait_for_threads(threads)
        results  = [qr.get() for qr in result_queues]
        return self.generate_result_from_layer_results(results,names)


    def generate_result_from_layer_results(self,results,names):
        return MeasureResult(results, names, self)

    def eval_samples_first(self,activations_iterator:ActivationsIterator, queues:List[Queue], inner_queues:List[Queue]):

        for activations, x_transformed in activations_iterator.samples_first():
            self.put_value(queues, x_transformed)
            self.put_values(inner_queues,activations)
            self.signal_iteration_end(inner_queues)
        self.signal_iteration_end(queues)


    def eval_transformations_first(self, activations_iterator: ActivationsIterator, queues: List[Queue],inner_queues: List[Queue]):

        for transformation, batch_activations in activations_iterator.transformations_first():
            self.put_value(queues,transformation)
            for x, batch_activation in batch_activations:
                self.put_values(inner_queues,batch_activation)
            self.signal_iteration_end(inner_queues)
        self.signal_iteration_end(queues)


class SamplesFirstPerLayerMeasure(PerLayerMeasure):
    def __init__(self):
        super().__init__(ActivationsOrder.SamplesFirst)

class TransformationsFirstPerLayerMeasure(PerLayerMeasure):
    def __init__(self):
        super().__init__(ActivationsOrder.TransformationsFirst)


from .base import NumpyMeasure
from ..measure import ActivationsByLayer
from .. import MeasureResult
from .activations_iterator import ActivationsIterator
import numpy as np


def divide_activations(num:ActivationsByLayer, den:ActivationsByLayer)->ActivationsByLayer:
    #TODO evaluate other implementations
    # https://stackoverflow.com/questions/26248654/how-to-return-0-with-divide-by-zero
    eps = 0

    measures = []  # coefficient of variations
    for num_values,den_values in zip(num, den):
        # print(layer_v_transformations.shape,layer_v_samples.shape)

        normalized_measure = num_values.copy()

        normalized_measure[den_values  > eps] /= den_values [den_values  > eps]
        both_below_eps = np.logical_and(den_values  <= eps,
                                        num_values <= eps)
        normalized_measure[both_below_eps] = 1

        only_baseline_below_eps = np.logical_and(
            den_values  <= eps,
            num_values > eps)
        # print("num", np.where(num_values > eps))
        # print("den", np.where(den_values <= eps))
        # print("both", np.where(only_baseline_below_eps))

        normalized_measure[only_baseline_below_eps] = np.inf
        measures.append(normalized_measure)
    return measures

class QuotientMeasure(NumpyMeasure):
    def __init__(self, numerator_measure:NumpyMeasure, denominator_measure:NumpyMeasure):
        super().__init__()
        self.numerator_measure=numerator_measure
        self.denominator_measure=denominator_measure

    def __repr__(self):
        return f"QM({self.numerator_measure}_DIV_{self.denominator_measure})"

    def eval(self,activations_iterator:ActivationsIterator,verbose=False)->MeasureResult:
        v_transformations = self.numerator_measure.eval(activations_iterator,verbose=False)
        v_samples=self.denominator_measure.eval(activations_iterator,verbose=False)
        v=divide_activations(v_transformations.layers, v_samples.layers)

        layer_names = activations_iterator.layer_names()
        return MeasureResult(v,layer_names,self)


